# -*- coding: UTF-8 -*-

import math
from typing import Tuple, Union, Literal

import numpy as np
from colormap import rgb2hex
from matplotlib.colors import ListedColormap, BoundaryNorm
from numpy import asarray
from pandas import DataFrame
from scipy import sparse
from ykenan_log import Logger

from ._constant_ import dense_data, sparse_data, sparse_matrix, matrix_data, number, collection

log = Logger("scLift_util_matrix")


def to_dense(sm: matrix_data, is_array: bool = False) -> dense_data:
    """
    Convert sparse matrix to dense matrix
    :param sm: sparse matrix
    :param is_array: True converted to array form, False natural output
    :return: dense matrix
    """

    if sm is None:
        log.warn("The input matrix (sm parameter) is not feasible")
        return np.asmatrix([])

    if isinstance(sm, sparse_data):
        return np.array(sm.todense()) if is_array else sm.todense()

    dense_sm = sm.todense() if sparse.issparse(sm) else sm
    return np.array(dense_sm) if is_array else dense_sm


def to_sparse(dm: dense_data, way_callback=sparse.csr_matrix, is_matrix: bool = True) -> sparse_matrix:
    """
    Convert dense matrix to sparse matrix
    :param dm: dense matrix
    :param way_callback: How to form sparse matrix
    :param is_matrix: True converted to matrix form, False natural output
    :return: sparse matrix
    """

    if dm is None:
        log.warn("The input matrix (dm parameter) is not feasible")
        return sparse.coo_matrix([])

    if isinstance(dm, dense_data):
        return way_callback(dm) if is_matrix else dm

    sparse_m = dm if sparse.issparse(dm) else way_callback(dm)
    return way_callback(sparse_m) if is_matrix else sparse_m


def sum_min_max(data: matrix_data, axis: int = 1) -> Tuple[number, number]:
    """
    Obtain the minimum/maximum sum of rows in the matrix
    :param data: matrix data
    :param axis: {0, 1} 1: row, 0: col
    :return: Minimum value of rows, maximum value of rows
    """
    rows, cols = data.shape

    if rows == 0 or cols == 0:
        return 0, 0

    rows_sum = list(np.array(data.sum(axis=axis)).flatten())
    return min(rows_sum), max(rows_sum)


def get_index(position: number, positions_list: list) -> Union[number, Tuple[number, number]]:
    """
    Search for position information. Similar to half search.
        If the position exists in the list, return the index.
        If it does not exist, return the index located between the two indexes
    :param position: position
    :param positions_list: position list
    :return: position index
    """
    # sort
    positions_list.sort()
    # search
    position_size: int = len(positions_list)
    left, right = 0, position_size - 1

    while left <= right:
        mid = (left + right) // 2

        if positions_list[mid] == position:
            return mid
        elif positions_list[mid] > position:
            right = mid - 1
        else:
            left = mid + 1

    return right, left


def list_duplicate_set(data: list) -> list:
    """
    Append numbering to duplicate information
    :param data:
    :return:
    """
    new_data = []
    range_data = range(len(data))

    for i in range_data:

        # judge duplicate
        if data[i] not in new_data:
            new_data.append(data[i])
        else:
            j: int = 2

            while True:

                if not isinstance(data[i], str):
                    log.warn("")

                # format new index
                elem: str = data[i] + "+" + str(j)

                # Add index
                if elem not in new_data:
                    new_data.append(elem)
                    break

                j += 1

    return new_data


def get_sub_data(data: collection, size: int) -> collection:
    # get information
    old_size = len(data)

    if size >= old_size:
        log.warn("The given size is greater than the original data size")
        return data

    old_data: list = data.copy()
    rate = size / old_size
    # add sub data
    new_data: list = []

    for i in range(size):
        new_data.append(old_data[math.floor(rate * i)])

    return new_data


def set_colors(data: collection, big_rgb_value: int = 150) -> Tuple:
    log.info("Start setting color gradient information")
    new_data: list = list(set(data)).copy()
    new_data.sort()
    size = len(new_data)
    # get zero location
    index = get_index(0.0, new_data)
    # get zero index
    middle_index = index if isinstance(index, number) else index[1]

    colors = []
    bounds = []

    # judge colors size and bounds size
    if size >= big_rgb_value * 2:

        # Add colors smaller than zero
        for i in range(big_rgb_value + 1):
            colors.append(rgb2hex(i, big_rgb_value, i))

        # Add colors greater than zero
        for j in range(1, big_rgb_value + 1):
            colors.append(rgb2hex(big_rgb_value, big_rgb_value - j, big_rgb_value - j))

        colors_size = len(colors)
        # Sampling data
        after_size = size - middle_index
        colors_size_half = colors_size // 2
        # add binds
        rate = middle_index / colors_size_half
        rate_after = after_size / (colors_size - colors_size_half)

        # Add data less than zero
        for m in range(colors_size_half):
            bounds.append(new_data[math.floor(rate * m)])

        # Add data greater than zero
        for n in range(colors_size - colors_size_half):
            if size > middle_index + math.floor(rate_after * n):
                bounds.append(new_data[middle_index + math.floor(rate_after * n)])
            else:
                break

        if bounds[-1] != new_data[-1]:
            bounds.append(new_data[-1])

    else:
        # Obtain colors
        size_half = size // 2
        rate_before = 150 / size_half
        rate_after = 150 / (size - size_half)

        # Add colors smaller than zero
        for i in range(size_half + 1):
            rgb = math.floor(rate_before * i)
            colors.append(rgb2hex(rgb, big_rgb_value, rgb))

        # Add colors greater than zero
        for j in range(1, size - size_half):
            rgb = math.floor(rate_after * j)
            colors.append(rgb2hex(big_rgb_value, big_rgb_value - rgb, big_rgb_value - rgb))

        # Data does not need to change
        bounds = new_data

    # The purpose of adding a value to each number is to ensure that each value falls within a certain color range, rather than the color edge
    new_bounds = [k - 1e-10 for k in bounds]
    new_bounds.append(bounds[-1] + 1e-5)

    # judge size
    colors_size = len(colors)
    bounds_size = len(new_bounds)

    if colors_size == bounds_size - 2:
        colors.append(rgb2hex(big_rgb_value, 0, 0))
    elif colors_size < bounds_size - 1:
        log.error(f"There are {bounds_size - 1} color bins including extensions, but colors = {colors_size}; colors must equal or exceed the number of bins")
        raise ValueError(f"There are {bounds_size - 1} color bins including extensions, but colors = {colors_size}; colors must equal or exceed the number of bins")

    # Return the data required for drawing
    cmap = ListedColormap(colors)
    norm = BoundaryNorm(new_bounds, len(colors))
    return cmap, norm


def split_matrix(data: matrix_data, axis: Literal[0, 1] = 0, chunk_number: int = 1000) -> list:
    # get size
    new_data = to_dense(data, is_array=True)
    rows, cols = new_data.shape

    # get number
    total: int = rows if axis == 0 else cols

    # get split number
    split_number = total // chunk_number

    # Determine whether to divide equally
    tail_number = total % chunk_number
    if tail_number != 0:
        split_number += 1

    log.info(f"Divide the matrix into {split_number} parts")

    # Add data
    split_data_list = []
    for i in range(split_number):
        # set index
        start_index: int = i * chunk_number
        end_index: int = total if (i + 1) * chunk_number > total else (i + 1) * chunk_number
        split_data = new_data[start_index:end_index, :] if axis == 0 else new_data[:, start_index:end_index]
        split_data_list.append(split_data)

    return split_data_list


def merge_matrix(datas: list, axis: Literal[0, 1] = 0) -> list:
    # get size
    size = len(datas)
    range_size = range(size)

    # get row col
    constant: int = datas[0].shape[1] if axis == 0 else datas[0].shape[0]
    total: int = 0
    shapes: list = []

    # get chunk size
    for i in range_size:
        shape = datas[i].shape
        judge = shape[1] if axis == 0 else shape[0]

        # judge size
        if judge != constant:
            log.error("")
            raise ValueError("")

        total += shape[0] if axis == 0 else shape[1]
        shapes.append(shape)

    # format matrix
    matrix_shape = (total if axis == 0 else constant, constant if axis == 0 else total)
    log.info(f"Merge matrix {matrix_shape} shape")
    matrix: matrix_data = np.zeros(matrix_shape)

    # merge matrix
    col_record: int = 0
    for i in range_size:
        col_i = datas[i].shape[0] if axis == 0 else datas[i].shape[1]
        if axis == 0:
            matrix[col_record:col_record + col_i, :] = datas[i]
        else:
            matrix[:, col_record:col_record + col_i] = datas[i]
        col_record += col_i

    return matrix


def list_index(data: list) -> Tuple[list, collection]:
    info: list = []

    size: int = len(data)

    types: asarray = np.unique(data)

    for type_ in types:
        type_info: list = []
        for i in range(size):
            if type_ == data[i]:
                type_info.append(i)
        info.append(set(type_info))

    return info, types


def numerical_bisection_step(min_value: float, max_value: float, step_length: float) -> Tuple[collection, int]:
    if min_value > max_value:
        log.error("")
        raise ValueError("")

    number_list: list = []

    i = 0
    while min_value <= max_value:
        number_list.append(min_value)
        i += 1
        min_value += step_length

    return number_list, i


def get_real_predict_label(
    df: DataFrame,
    map_cluster: Union[str, collection],
    clusters: str = "clusters",
    value: str = "value"
) -> Tuple[DataFrame, int, list]:
    df_sort: DataFrame = df.sort_values([value], ascending=False)

    # Obtain the type of positive set clustering corresponding to the trait
    cluster_list: list = []
    if isinstance(map_cluster, str):
        cluster_list.append(map_cluster)
    else:
        cluster_list = list(map_cluster)

    # total label size
    total_size = df.shape[0]

    # true label size
    df_sort.insert(0, "true_label", 0)
    df_sort.loc[df_sort[df_sort[clusters].isin(cluster_list)].index, "true_label"] = 1

    # predict label size
    df_cluster = df[df[clusters].isin(cluster_list)].copy()
    df_cluster_size = df_cluster.shape[0]
    predict_label = list(np.ones(df_cluster_size))
    predict_label.extend(np.zeros(total_size - df_cluster_size))
    df_sort.insert(0, "predict_label", 0)
    df_sort.loc[:, "predict_label"] = predict_label
    df_sort["predict_label"] = df_sort["predict_label"].astype(int)

    return df_sort, df_cluster_size, cluster_list


def strings_map_numbers(str_list: list, start: int = 0) -> list:
    # Create an empty dictionary to store the mapping of strings to numerical values
    mapping = {}

    # Traverse the list and assign a unique numerical value to each string
    for i, item in enumerate(set(str_list), start=start):  # 使用set去除重复项，并从1开始编号
        mapping[item] = i

    # Use list derivation to convert strings to corresponding numerical values
    numeric_list = [mapping[item] for item in str_list]

    return numeric_list
